@@ -1,6 +1,5 @@
 /* util.c - Several utility routines for cpio.
-   Copyright (C) 1990-1992, 2001, 2004, 2006-2007, 2010-2011, 2014-2015,
-   2017 Free Software Foundation, Inc.
+   Copyright (C) 1990-2023 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -60,8 +59,8 @@ tape_empty_output_buffer (int out_des)
   static long output_bytes_before_lseek = 0;
 
   /* Some tape drivers seem to have a signed internal seek pointer and
-     they lose if it overflows and becomes negative (e.g. when writing 
-     tapes > 2Gb).  Doing an lseek (des, 0, SEEK_SET) seems to reset the 
+     they lose if it overflows and becomes negative (e.g. when writing
+     tapes > 2Gb).  Doing an lseek (des, 0, SEEK_SET) seems to reset the
      seek pointer and prevent it from overflowing.  */
   if (output_is_special
      && ( (output_bytes_before_lseek += output_size) >= 1073741824L) )
@@ -79,8 +78,7 @@ tape_empty_output_buffer (int out_des)
 
       if (output_is_special
 	  && (bytes_written >= 0
-	      || (bytes_written < 0
-		  && (errno == ENOSPC || errno == EIO || errno == ENXIO))))
+	      || (errno == ENOSPC || errno == EIO || errno == ENXIO)))
 	{
 	  get_next_reel (out_des);
 	  if (bytes_written > 0)
@@ -106,7 +104,7 @@ static ssize_t sparse_write (int fildes, char *buf, size_t nbyte, bool flush);
    descriptor OUT_DES and reset `output_size' and `out_buff'.
    If `swapping_halfwords' or `swapping_bytes' is set,
    do the appropriate swapping first.  Our callers have
-   to make sure to only set these flags if `output_size' 
+   to make sure to only set these flags if `output_size'
    is appropriate (a multiple of 4 for `swapping_halfwords',
    2 for `swapping_bytes').  The fact that DISK_IO_BLOCK_SIZE
    must always be a multiple of 4 helps us (and our callers)
@@ -188,8 +186,8 @@ tape_fill_input_buffer (int in_des, int num_bytes)
 {
 #ifdef BROKEN_LONG_TAPE_DRIVER
   /* Some tape drivers seem to have a signed internal seek pointer and
-     they lose if it overflows and becomes negative (e.g. when writing 
-     tapes > 4Gb).  Doing an lseek (des, 0, SEEK_SET) seems to reset the 
+     they lose if it overflows and becomes negative (e.g. when writing
+     tapes > 4Gb).  Doing an lseek (des, 0, SEEK_SET) seems to reset the
      seek pointer and prevent it from overflowing.  */
   if (input_is_special
       && ( (input_bytes_before_lseek += num_bytes) >= 1073741824L) )
@@ -332,8 +330,8 @@ tape_buffered_peek (char *peek_buf, int in_des, int num_bytes)
 
 #ifdef BROKEN_LONG_TAPE_DRIVER
   /* Some tape drivers seem to have a signed internal seek pointer and
-     they lose if it overflows and becomes negative (e.g. when writing 
-     tapes > 4Gb).  Doing an lseek (des, 0, SEEK_SET) seems to reset the 
+     they lose if it overflows and becomes negative (e.g. when writing
+     tapes > 4Gb).  Doing an lseek (des, 0, SEEK_SET) seems to reset the
      seek pointer and prevent it from overflowing.  */
   if (input_is_special
       && ( (input_bytes_before_lseek += num_bytes) >= 1073741824L) )
@@ -404,7 +402,7 @@ tape_toss_input (int in_des, off_t num_bytes)
 
       if (crc_i_flag && only_verify_crc_flag)
 	{
- 	  int k;
+	  int k;
 	  for (k = 0; k < space_left; ++k)
 	    crc += in_buff[k] & 0xff;
 	}
@@ -416,14 +414,14 @@ tape_toss_input (int in_des, off_t num_bytes)
 }
 
 void
-write_nuls_to_file (off_t num_bytes, int out_des, 
-                    void (*writer) (char *in_buf, int out_des, off_t num_bytes))
+write_nuls_to_file (off_t num_bytes, int out_des,
+		    void (*writer) (char *in_buf, int out_des, off_t num_bytes))
 {
   off_t	blocks;
   off_t	extra_bytes;
   off_t	i;
   static char zeros_512[512];
-  
+
   blocks = num_bytes / sizeof zeros_512;
   extra_bytes = num_bytes % sizeof zeros_512;
   for (i = 0; i < blocks; ++i)
@@ -603,7 +601,7 @@ create_all_directories (char const *name)
   char *dir;
 
   dir = dir_name (name);
-  
+
   if (dir == NULL)
     error (PAXEXIT_FAILURE, 0, _("virtual memory exhausted"));
 
@@ -637,9 +635,9 @@ create_all_directories (char const *name)
 void
 prepare_append (int out_file_des)
 {
-  int start_of_header;
-  int start_of_block;
-  int useful_bytes_in_block;
+  off_t start_of_header;
+  off_t start_of_block;
+  size_t useful_bytes_in_block;
   char *tmp_buf;
 
   start_of_header = last_header_start;
@@ -697,8 +695,8 @@ inode_val_compare (const void *val1, const void *val2)
   const struct inode_val *ival1 = val1;
   const struct inode_val *ival2 = val2;
   return ival1->inode == ival2->inode
-         && ival1->major_num == ival2->major_num
-         && ival1->minor_num == ival2->minor_num;
+	 && ival1->major_num == ival2->major_num
+	 && ival1->minor_num == ival2->minor_num;
 }
 
 static struct inode_val *
@@ -706,10 +704,10 @@ find_inode_val (ino_t node_num, unsigned long major_num,
 		 unsigned long minor_num)
 {
   struct inode_val sample;
-  
+
   if (!hash_table)
     return NULL;
-  
+
   sample.inode = node_num;
   sample.major_num = major_num;
   sample.minor_num = minor_num;
@@ -734,7 +732,7 @@ add_inode (ino_t node_num, char *file_name, unsigned long major_num,
 {
   struct inode_val *temp;
   struct inode_val *e = NULL;
-  
+
   /* Create new inode record.  */
   temp = (struct inode_val *) xmalloc (sizeof (struct inode_val));
   temp->inode = node_num;
@@ -846,11 +844,9 @@ get_next_reel (int tape_des)
   FILE *tty_out;		/* File for interacting with user.  */
   int old_tape_des;
   char *next_archive_name;
-  dynamic_string new_name;
+  dynamic_string new_name = DYNAMIC_STRING_INITIALIZER;
   char *str_res;
 
-  ds_init (&new_name, 128);
-
   /* Open files for interactive communication.  */
   tty_in = fopen (TTY_NAME, "r");
   if (tty_in == NULL)
@@ -925,7 +921,7 @@ get_next_reel (int tape_des)
     error (PAXEXIT_FAILURE, 0, _("internal error: tape descriptor changed from %d to %d"),
 	   old_tape_des, tape_des);
 
-  free (new_name.ds_string);
+  ds_free (&new_name);
   fclose (tty_in);
   fclose (tty_out);
 }
@@ -1007,7 +1003,7 @@ buf_all_zeros (char *buf, int bufsize)
 
 /* Write NBYTE bytes from BUF to file descriptor FILDES, trying to
    create holes instead of writing blockfuls of zeros.
-   
+
    Return the number of bytes written (including bytes in zero
    regions) on success, -1 on error.
 
@@ -1027,7 +1023,7 @@ sparse_write (int fildes, char *buf, size_t nbytes, bool flush)
 
   enum { begin, in_zeros, not_in_zeros } state =
 			   delayed_seek_count ? in_zeros : begin;
-  
+
   while (nbytes)
     {
       size_t rest = nbytes;
@@ -1042,7 +1038,7 @@ sparse_write (int fildes, char *buf, size_t nbytes, bool flush)
 	      if (state == not_in_zeros)
 		{
 		  ssize_t bytes = buf - start_ptr + rest;
-		  
+
 		  n = write (fildes, start_ptr, bytes);
 		  if (n == -1)
 		    return -1;
@@ -1091,8 +1087,8 @@ sparse_write (int fildes, char *buf, size_t nbytes, bool flush)
       if (n != 1)
 	return n;
       delayed_seek_count = 0;
-    }      
-  
+    }
+
   return nwritten + seek_count;
 }
 
@@ -1140,8 +1136,16 @@ stat_to_cpio (struct cpio_file_stat *hdr, struct stat *st)
   hdr->c_nlink = st->st_nlink;
   hdr->c_uid = CPIO_UID (st->st_uid);
   hdr->c_gid = CPIO_GID (st->st_gid);
-  hdr->c_rdev_maj = major (st->st_rdev);
-  hdr->c_rdev_min = minor (st->st_rdev);
+  if (S_ISBLK (st->st_mode) || S_ISCHR (st->st_mode))
+    {
+      hdr->c_rdev_maj = major (st->st_rdev);
+      hdr->c_rdev_min = minor (st->st_rdev);
+    }
+  else
+    {
+      hdr->c_rdev_maj = 0;
+      hdr->c_rdev_min = 0;
+    }
   hdr->c_mtime = st->st_mtime;
   hdr->c_filesize = st->st_size;
   hdr->c_chksum = 0;
@@ -1222,7 +1226,7 @@ set_perms (int fd, struct cpio_file_stat *header)
   if (!no_chown_flag)
     {
       uid_t uid = CPIO_UID (header->c_uid);
-      gid_t gid = CPIO_GID (header->c_gid); 
+      gid_t gid = CPIO_GID (header->c_gid);
       if ((fchown_or_chown (fd, header->c_name, uid, gid) < 0)
 	  && errno != EPERM)
 	chown_error_details (header->c_name, uid, gid);
@@ -1239,13 +1243,13 @@ set_file_times (int fd,
 		const char *name, unsigned long atime, unsigned long mtime)
 {
   struct timespec ts[2];
-  
+
   memset (&ts, 0, sizeof ts);
 
   ts[0].tv_sec = atime;
   ts[1].tv_sec = mtime;
 
-  /* Silently ignore EROFS because reading the file won't have upset its 
+  /* Silently ignore EROFS because reading the file won't have upset its
      timestamp if it's on a read-only filesystem. */
   if (fdutimens (fd, name, ts) < 0 && errno != EROFS)
     utime_error (name);
@@ -1297,7 +1301,7 @@ cpio_safer_name_suffix (char *name, bool link_target, bool absolute_names,
 
 /* This is a simplified form of delayed set_stat used by GNU tar.
    With the time, both forms will merge and pass to paxutils
-   
+
    List of directories whose statuses we need to extract after we've
    finished extracting their subsidiary files.  If you consider each
    contiguous subsequence of elements of the form [D]?[^D]*, where [D]
@@ -1415,7 +1419,7 @@ cpio_mkdir (struct cpio_file_stat *file_hdr, int *setstat_delayed)
 {
   int rc;
   mode_t mode = file_hdr->c_mode;
-  
+
   if (!(file_hdr->c_mode & S_IWUSR))
     {
       rc = mkdir (file_hdr->c_name, mode | S_IWUSR);
@@ -1438,10 +1442,10 @@ cpio_create_dir (struct cpio_file_stat *file_hdr, int existing_dir)
 {
   int res;			/* Result of various function calls.  */
   int setstat_delayed = 0;
-  
+
   if (to_stdout_option)
     return 0;
-  
+
   /* Strip any trailing `/'s off the filename; tar puts
      them on.  We might as well do it here in case anybody
      else does too, since they cause strange things to happen.  */
@@ -1530,7 +1534,7 @@ arf_stores_inode_p (enum archive_format arf)
     }
   return 1;
 }
-  
+
 void
 cpio_file_stat_init (struct cpio_file_stat *file_hdr)
 {
